#include "dcdialog.CH"
// #include "sql.ch"
// #include "sqlext.ch"
#INCLUDE "fileio.CH"
#INCLUDE "appevent.CH"
#INCLUDE "dmlb.CH"

#define __ERR_LOG_FILE    'SQLERROR.LOG'
#define __CRLF            (Chr(13)+Chr(10))

#define SQL_XPP_SUCCESS           (-1)
#define SQL_XPP_ERROR             (-2)
#define SQL_XPP_WARNING           (-3)

/* Options for SQLDriverConnect */
#define SQL_DRIVER_NOPROMPT             0
#define SQL_DRIVER_COMPLETE             1
#define SQL_DRIVER_PROMPT               2
#define SQL_DRIVER_COMPLETE_REQUIRED    3

/* SQL_CURSOR_TYPE options */
#define SQL_CURSOR_FORWARD_ONLY         0
#define SQL_CURSOR_KEYSET_DRIVEN        1
#define SQL_CURSOR_DYNAMIC              2
#define SQL_CURSOR_STATIC               3
#define SQL_CURSOR_TYPE_DEFAULT         SQL_CURSOR_FORWARD_ONLY // Default

/* SQL_CONCURRENCY options */
#define SQL_CONCUR_READ_ONLY            1
#define SQL_CONCUR_LOCK                 2
#define SQL_CONCUR_ROWVER               3
#define SQL_CONCUR_VALUES               4
#define SQL_CONCUR_DEFAULT              SQL_CONCUR_READ_ONLY //

/* SQL_TXN_ISOLATION_OPTION bitmasks */
#define SQL_TXN_READ_UNCOMMITTED            0x00000001 //L
#define SQL_TXN_READ_COMMITTED              0x00000002 //L
#define SQL_TXN_REPEATABLE_READ             0x00000004 //L
#define SQL_TXN_SERIALIZABLE                0x00000008 //L

#define SQL_DEFAULT_TXN_ISOLATION           26

#xtranslate NTrim(<n>)   =>  LTrim(Str(<n>))
#define CRLF  Chr(13)+Chr(10)

* --------------

FUNCTION DC_SQLConnection( nType )

LOCAL i, aRef, aData, cDSN, cUid, cPwd, nThreadID := ThreadID(), ;
      cString, aTokens, cConnect, lDestroy, oConn, oSQL

STATIC saConnection[0]

IF !IsFunction("SQLConnection")
  DllLoad('SQLXPP.DLL')
ENDIF

IF !IsFunction("SQLConnection")
  DC_WinAlert('Cannot load SQLXPP.DLL')
  RETURN nil
ENDIF

lDestroy := (Valtype(nType) == 'L')

DEFAULT nType := DCSQL_CONNECT_TYPE_READWRITE

IF Len(saConnection) < nThreadID
   ASize(saConnection, nThreadID)
ENDIF

IF saConnection[nThreadID] == NIL
  saConnection[nThreadID] := Array(2)
ENDIF

IF lDestroy
   IF Valtype(saConnection[nThreadID,1]) == 'O'
     saConnection[nThreadId,1]:destroy()
   ENDIF
   IF Valtype(saConnection[nThreadID,2]) == 'O'
     saConnection[nThreadId,2]:destroy()
   ENDIF
   saConnection[nThreadId] := Nil
   RETURN nil
ELSEIF Valtype(saConnection[nThreadID,nType]) == 'O'
   oConn := saConnection[nThreadID,nType]
   RETURN oConn
ENDIF

cConnect := DC_SQLConnectString()

// establish the ODBC connection
oSQL := &('SQLConnection()')
oConn := oSQL:new(,,,nType == 2 )

IF nType == DCSQL_CONNECT_TYPE_READONLY
  oConn:transactionIsolation := SQL_TXN_READ_UNCOMMITTED
ENDIF

// put all read only connections into a different static array than read/write connections

saConnection[nThreadID,nType] := oConn

IF cConnect = 'DSN'
  aTokens := DC_TokenArray(cConnect,';')
  FOR i := 1 TO Len(aTokens)
    cString := Upper(aTokens[i])
    IF cString='DSN'
      cDSN := SubStr(cString,At('=',cString)+1)
    ELSEIF cString='UID'
      cUid := SubStr(cString,At('=',cString)+1)
    ELSEIF cString='PWD'
      cPwd := SubStr(cString,At('=',cString)+1)
    ENDIF
  NEXT
  oConn:connect( cDSN, cUid, cPwd )
ELSE
  oConn:driverConnect(nil, @cConnect, SQL_DRIVER_NOPROMPT)
ENDIF

if !oConn:isConnected
  oConn:destroy()
  saConnection[nThreadId] := nil
  oConn := nil
  SQLErrorLog("= ODBC ERROR ==========" + __CRLF +;
         "Date: " + DTOS(Date()) + ", Time: " + Time() + __CRLF +;
         "App: " + AppName(.t.) + ", Thread: " + LTrim(Str(nThreadID)) + __CRLF +;
         "Unable to connect to ODBC data source!" + __CRLF)
  IF nThreadID > 1
    ThreadObject():quit()
  ENDIF
else
   // it's very important that all SQLConnection objects are explicity :destroyed() when no
   // longer needed. However in case the programmer forgets to do so, or if the thread crashes,
   // one way to make sure this connection is automatically destroyed at thread termination
   // is to make use of the thread :atEnd ivar:

  ThreadObject():atEnd := {||DC_SQLConnection(.t.)}
endif

oConn:DateTimeAsDate := .t.
oConn:onError := {|oOwner,cError,nRetCode,aErrObj| ;
                   (dcbdebug oConn), ;
                   DC_SQLErrorHandler(oOwner,cError,nRetCode,aErrObj)}

RETURN oConn

//-----------------------------------------------------------------------------
// The following SQL error handling function will not only log all errors to a file
// but it will also attempt to automatically reconnect to the data source when it
// senses that a connection has been lost.
//
PROCEDURE DC_SQLErrorHandler( oOwner, cError, nRetCode, aErrObj )

Local oConn

if oOwner:isDerivedFrom('SQLConnection')
   oConn := oOwner
else
   oConn := oOwner:Connection
   if oOwner:isDerivedFrom('SQLStatement') .and. ValType(oOwner:SQLString) == 'C'
      cError += __CRLF + oOwner:SQLString
   endif
endif

SQLErrorLog("= SQL ERROR ==========" + __CRLF +;
   "Date: " + DTOS(Date()) + ", Time: " + Time() + __CRLF +;
   "App: " + AppName(.t.) + ", Thread: " + LTrim(Str(ThreadID())) + __CRLF +;
   cError + __CRLF)

if AScan(aErrObj, {|o|Left(o:SQLState,2)=='08'}) > 0 // communications link failure

   oConn:ShowConnErrors := .f.                  // prevent recursion in subsequent errors
   SQLErrorLog("*** ATTEMPTING TO RECONNECT ***" + __CRLF )
   Sleep(300)
   while ! oConn:Reconnect(.t.)                 // keep trying until connected
      Sleep(900)
   end
   SQLErrorLog("*** RECONNECTED SUCCESSFULLY *** Date: "+DTOS(Date())+", Time: "+Time()+__CRLF +;
      "App: " + AppName(.t.) + ", Thread: " + LTrim(Str(ThreadID())) + __CRLF )
   if oOwner:isDerivedFrom('SQLStatement')

      // ** NOTE **
      // Re-execute the statement that just failed due to the comm link failure
      // For simple SQL statemens this is just fine, the program should continue as if nothing happened.
      // However if the statement produces a result set (SELECT), then you may need to do additional work
      // since re-executing the statement will position the row pointer at the top of the result set.

      oOwner:Execute( oOwner:ParamList )
   endif
   oConn:ShowConnErrors := .t.

endif

Return

//-----------------------------------------------------------------------------
// Write a string to a log file.
//
STATIC procedure SQLErrorLog( cText )

Local i := 0, nHandle

if ValType(cText) != "C" .or. Empty(cText)
  Return
endif

if File(__ERR_LOG_FILE)
   while (nHandle := FOpen(__ERR_LOG_FILE, FO_WRITE + FO_DENYWRITE)) < 0
      if ++i > 9
         Return
      endif
      Sleep(20)
   end
   FSeek(nHandle, 0, FS_END)                    // position to end of file
else
   while (nHandle := FCreate(__ERR_LOG_FILE)) < 0
      if ++i > 9
         Return
      endif
      Sleep(20)
   end
endif

FWrite(nHandle, cText + __CRLF)
FClose(nHandle)

Return

* ------------------

FUNCTION DC_SQLCursor( oConn, i, nSuccess, cStatement, bSQLParam, aPrimaryKeys, ;
                       nMaxRows, abAddCol, bFilter, lTrim, lDateTimeAsDate, ;
                       cTableName )

Local oCursor, oSql, nType, j, acParams

DEFAULT nSuccess := SQL_XPP_ERROR, ;
        nMaxRows := 30000, ;
        lTrim := .T., ;
        lDateTimeAsDate := .T., ;
        i := 0

IF i == 1
  nType := DCSQL_CONNECT_TYPE_READONLY
ELSE
  nType := DCSQL_CONNECT_TYPE_READWRITE
ENDIF

IF Valtype(oConn) # 'O'
  oConn := DC_SQLConnection( nType )
ENDIF

IF Valtype(oConn) # 'O' .OR. !oConn:isConnected
  DC_WinAlert("Connection error!")
  Return nil
endif

oConn:displayErrors := .t.

IF i # 2 .AND. !Empty(bSQLParam) .AND. '?' $ cStatement
  IF Valtype(bSQLParam) == 'B'
    acParams := Eval(bSQLParam)
  ELSE
    acParams := bSQLParam
  ENDIF
  IF Valtype(acParams) # 'A'
    acParams := { acParams }
  ENDIF
  FOR j := 1 TO Len(acParams)
    cStatement := Strtran( cStatement, '?', DC_Var2SQL(acParams[j]),,1 )
  NEXT
ENDIF

do case

   case i == 0
      oSql := &('SQLSelect()')
      oCursor := oSql:new( cStatement, oConn, nil, SQL_CURSOR_DYNAMIC )
      nSuccess := oCursor:execute()

   case i == 1
      oSql := &('SQLSelect()')
      oCursor := oSql:new(cStatement, oConn, SQL_CONCUR_READ_ONLY, SQL_CURSOR_DYNAMIC)
      nSuccess := oCursor:execute()

   case i == 2
      oSql  := &('SQLDataSet()')
      oCursor := oSql:new(cStatement, oConn, bSQLParam, aPrimaryKeys, ;
                          nMaxRows, abAddCol, bFilter, lTrim, lDateTimeAsDate)
      nSuccess := oCursor:execute() // don't really need to execute an SQLDataSet, it's just here for consistency

   case i == 3
      oSql  := &('SQLListColumnPrivileges()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute(,,cTableName)

   case i == 4
      oSql := &('SQLListColumns()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute(,,cTableName)

   case i == 5
      oSql := &('SQLListForeignKeys()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute(,,,,,cTableName)

   case i == 6
      oSql := &('SQLListPrimaryKeys()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute(,,cTableName)

   case i == 7
      oSql := &('SQLListProcedureColumns()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute()

   case i == 8
      oSql := &('SQLListProcedures()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute()

   case i == 9
      oSql := &('SQLListSpecialColumns()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute(,,,cTableName)

   case i == 10
      oSql := &('SQLListStatistics()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute(,,cTableName)

   case i == 11
      oSql := &('SQLListTablePrivileges()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute()

   case i == 12
      oSql := &('SQLListTables()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute()

   case i == 13
      oSql := &('SQLListTypeInfo()')
      oCursor := oSql:new( oConn )
      nSuccess := oCursor:execute()

endcase

oConn:displayErrors := .f.

if nSuccess == SQL_XPP_ERROR
  Return nil
endif

Return oCursor

* --------------------

FUNCTION DC_SQLFieldBlock( oCursor, nField )

RETURN {|x| iif(PCount()==0, oCursor:fieldGet(nField), ;
                             oCursor:fieldPut(nField, x))}

* ---------------

FUNCTION DC_SQLGetSetBlock( oCursor, nField )

Local oCol, nLen, nDec

oCol := oCursor:getSQLColumn(nField)
nLen := oCol:length

IF oCol:ValType == "C"

  Return {|x|if(x==NIL,PadR(oCursor:fieldGet(nField),nLen),;
              oCursor:fieldPut(nField,Trim(x)))}

ELSEIF oCol:ValType == "N"

  nDec := oCol:decimals
  Return {|x|if(x==NIL,oCursor:fieldGet(nField),;
             oCursor:fieldPut(nField,x))}

ENDIF

RETURN {|x|if(x==NIL,oCursor:fieldGet(nField),;
           oCursor:fieldPut(nField,x))}

* --------------

FUNCTION DC_SQLSortBlock( oBrowse, nField, nColumn )

RETURN {|a,b,descend|;
   IIF(valtype(nField)=='C',nField:=oBrowse:dataSource:fieldPos(nField),nil), ;
       oBrowse:dataSource:sort(IIF(descend,;
       {|x,y|IIF(Valtype(x[nField])='C', ;
             Upper(x[nField]) > Upper(y[nField]), x[nField] > y[nField])}, ;
       {|x,y|IIF(Valtype(x[nField])='C', ;
             Upper(x[nField]) < Upper(y[nField]), x[nField] < y[nField])})), ;
       nColumn := nField}

* --------------

FUNCTION DC_SQLDataSet( oDataSet, bFilter )

LOCAL i, oNewDataSet, nRecNo := oDataSet:recNo()

oDataSet:clearRelation()
oDataSet:goTo(1)

IF !Empty(bFilter)
  oDataSet:locate( bFilter )
ENDIF
oNewDataSet := &('SQLDataSet():new()')

FOR i := 1 TO oDataSet:fCount()
  oNewDataSet:addColumn(oDataSet:fieldName(i))
NEXT

DO WHILE !oDataSet:eof()

  IF bFilter == NIL .OR. Eval(bFilter,oDataSet:recordGet())
    oNewDataSet:append( oDataSet:recordGet() )
  ENDIF

  oDataSet:skip()

ENDDO

oDataSet:goto(nRecNo)
oNewDataSet:goto(1)

RETURN oNewDataSet

* --------------

/*
#if XPPVER > 1900000
  FUNCTION SQLExpressKey(); Return 6341671.5
#else
  FUNCTION SQLExpressKey(); Return 9675932601
#endif
*/

* ---------------

FUNCTION DC_SQLConnectString( cConnect )

STATIC scConnect := nil

LOCAL aRef, aData, cOldConnect := scConnect

IF Empty(cConnect) .AND. !Empty(scConnect)
  RETURN scConnect
ENDIF

IF Empty(scConnect) .AND. Empty(cConnect)
  cConnect := 'DCSQL.INI'
ENDIF

IF Right(Upper(cConnect),4) == '.INI'
  aRef := {'ConnStr'}
  aData := { Space(200) }

  DC_INILOAD(cConnect,'DATABASE',aData,aRef)
  scConnect := Trim(aData[1])

ELSEIF !Empty(cConnect)
  scConnect := cConnect
ENDIF

RETURN scConnect

* ------------

FUNCTION DC_Var2SQL( xVal )

LOCAL cValtype := Valtype(xVal), cYear, cMonth, cDay

IF cValtype == 'C'
   IF "'" $ xVal
      return "'" + strtran(xVal, "'", "''") + "'"
   ENDIF
   return "'" + xVal + "'"
ELSEIF cValtype == 'D'
   IF Empty(xVal)
      xVal := Ctod('01/01/1900')
   ENDIF
   cYear := Alltrim(Str(Year(xVal)))
   cMonth := Strtran(Str(Month(xVal),2,0),' ','0')
   cDay := Strtran(Str(Day(xVal),2,0),' ','0')
   RETURN "'" + cYear + '-' + cMonth + '-' + cDay + "'"
   //{ d '2004-01-01' }
ELSEIF cValtype == 'N'
   RETURN Ltrim(Str(xVal))
ELSEIF cValtype == 'L'
   IF xVal
     RETURN 'TRUE'
   ENDIF
   RETURN 'FALSE'
ELSEIF cValtype == 'U'
   RETURN 'NULL'
ENDIF

RETURN "'" + Trim(DC_XtoC(xVal)) + "'"

* ---------------

FUNCTION DC_Array2SQLInsert( aData, cTable, aExclude )

LOCAL cStatement := "", i, j, aFields, cValue, aResults, aIvars, ;
      oRecord, cFieldName

DEFAULT cTable := Alias()

IF !Empty(aData)

  cStatement := 'BEGIN TRANSACTION;' + Chr(13)+Chr(10)

  FOR i := 1 TO Len(aData)

    aFields := aData[i]

    IF Valtype(aFields) == 'O'

      cStatement += 'INSERT INTO ' + cTable + ' ('

      oRecord := aFields
      aResults := oRecord:classDescribe()
      aIvars := {}
      FOR j := 1 TO Len(aResults[3])
        AAdd(aIvars,Alltrim(aResults[3,j,1]))
      NEXT

      FOR j := 1 TO Len(aIvars)
        cFieldName := aIvars[j]
        IF Upper(cFieldName) == '__NOIVAR'
          LOOP
        ELSEIF Valtype(aExclude) == 'A' .AND. AScan(aExclude,{|c|Upper(c)==Upper(cFieldName)}) > 0
          LOOP
        ENDIF
        IF IsMemberVar(oRecord,cFieldName)
          IF Upper(cFieldName) == 'USER'
            cFieldName := '[USER]'
          ENDIF
          cStatement += cFieldName + IIF(j<Len(aIvars),',','')
        ENDIF
      NEXT

      cStatement += ') VALUES ('

      FOR j := 1 TO Len(aIvars)
        cFieldName := aIvars[j]
        IF Upper(cFieldName) == '__NOIVAR'
          LOOP
        ELSEIF Valtype(aExclude) == 'A' .AND. AScan(aExclude,{|c|Upper(c)==Upper(cFieldName)}) > 0
          LOOP
        ENDIF
        IF IsMemberVar(oRecord,cFieldName)
          cValue := DC_Var2SQL(oRecord:&(cFieldName))
          cStatement += cValue + IIF(j<Len(aIvars),',','')
        ENDIF
      NEXT

    ELSEIF Valtype(aFields) == 'A'

      cStatement += 'INSERT INTO ' + cTable + ' VALUES ('

      FOR j := 1 TO Len(aFields)
        cValue := DC_Var2SQL(aFields[j])
        cStatement += cValue + IIF(j<Len(aFields),',','')
      NEXT

    ENDIF

    cStatement += ');' + Chr(13) + Chr(10)

  NEXT

  cStatement += 'COMMIT WORK;' + Chr(13) + Chr(10)

ENDIF

RETURN cStatement

* ------------

FUNCTION DC_SQLInsert( cTable, aData )

LOCAL cStatement := "", i, j, aFields, cValue, aValues

DEFAULT cTable := Alias()

cStatement := 'BEGIN TRANSACTION;' + Chr(13)+Chr(10)

FOR i := 1 TO Len(aData)

  aFields := aData[i,1]
  aValues := aData[i,2]

  cStatement += 'INSERT INTO ' + cTable + ' ('

  FOR j := 1 TO Len(aFields)
    cStatement += aFields[j] + IIF(j<Len(aFields),',','')
  NEXT

  cStatement += ') VALUES ('

  FOR j := 1 TO Len(aValues)
    cValue := DC_Var2SQL(aValues[j])
    cStatement += cValue + IIF(j<Len(aValues),',','')
  NEXT

  cStatement += ');' + Chr(13) + Chr(10)

NEXT

cStatement += 'COMMIT WORK;' + Chr(13) + Chr(10)

RETURN cStatement

* -----------

FUNCTION DC_ApplySQLParams( cSqlStatement, aParams )

LOCAL i, cValue, cDate, xValue, nError

FOR i := 1 TO Len(aParams)
   xValue := aParams[i]
   IF Valtype(xValue) == 'N'
      cValue := Alltrim(Str(xValue))
   ELSEIF Valtype(xValue) == 'L'
      IF xValue
         cValue := 'TRUE'
      ELSE
         cValue := 'FALSE'
      ENDIF
   ELSEIF Valtype(xValue) == 'D'
      cDate := DtoS(xValue)
      cDate := Ltrim(Rtrim(SubStr(cDate,1,4) + '-' + ;
         SubStr(cDate,5,2) + '-' + ;
         SubStr(cDate,7,2)))
      cValue := "'" + cDate + "'"
   ELSEIF Valtype(xValue) == 'C'
      IF Left(xValue,1) = '[' .AND. Right(xValue,1) = ']' // literal (scalar functions)
        cValue := Substr(xValue,2,Len(xValue)-2)
        cValue := Strtran(cValue,'PARTS->','')
      ELSEIF "'" $ xValue
        cValue := "'" + Strtran(xValue,"'","''") + "'"
      ELSE
        cValue := "'" + xValue + "'"
      ENDIF
   ELSEIF xValue == NIL
      cValue := "''"
   ELSE
      cValue := Trim(DC_XtoC(xValue))
   ENDIF
   cSqlStatement := StrTran(cSqlStatement,'?',cValue,,1)
NEXT

RETURN cSqlStatement

* ------------

FUNCTION DC_GetSqlStatement( cTableName )

LOCAL cSqlStatement := '', cSqlTableName, cStatement, nFound

cSqlTableName := dbInfo( DBO_FILENAME )

IF cSqlTableName = 'SELECT '
  cSqlStatement := cSqlTableName
ELSE
  cSqlStatement := DC_AdsGetSqlStatement()
ENDIF

cStatement := Upper(cSqlStatement) + ' '
cStatement := Strtran(cStatement,Chr(13)+Chr(10),' ')

IF ' FROM ' $ cStatement
  DO WHILE ' FROM  ' $ cStatement
    cStatement := Strtran( cStatement, ' FROM  ', ' FROM ')
  ENDDO
  nFound := At( ' FROM ', cStatement)
  IF nFound > 0
    cStatement := SubStr(cStatement, nFound + 6)
    cTableName := Alltrim(Substr(cStatement,1,At(' ',cStatement)-1))
  ENDIF
ENDIF

RETURN cSqlStatement

* --------------

FUNCTION DC_GetSqlTablename()

LOCAL cTablename := ''

DC_GetSqlStatement( @cTableName )

RETURN cTableName
